#!/bin/bash
set -e

#
# Configure the databases for the AWS demo (creates the schema and
# tables needed).
#
# Waits for each database to become available by running a simple "SELECT READY" command.
# In the case of Oracle, as sqlplus is available before the database has started,
# it checks for the instance to be OPEN as well.
#
# Once the database is available, the DB-specific initialization script is copied
# to the containers and is run.
#

DIR=$(dirname "$0")

wait_for_ready() {
  attempts=1
  RESULT=$(eval "$cmd" | grep READY || true)
  while [[ -z ${RESULT} ]]; do
      ((attempts++))
      if [[ attempts -gt 30 ]]; then
          echo "Max wait time exceeded waiting for database to start"
          exit 1
      fi

    echo 'Sleeping for 5 seconds ... '
    sleep 5s
    RESULT=$(eval "$cmd" | grep READY || true)
  done
}

# North America
echo "Initializing North America - PostgreSQL"
cmd="kubectl exec deployments/north-america-postgres-db --namespace=default -- psql -U postgres -c \"SELECT 'READY'\""
wait_for_ready
pod=$(kubectl --namespace=default get pods -o jsonpath="{.items[0].metadata.name}" --selector=app=north-america-postgres-db)
kubectl cp "$DIR"/postgresql.sql "$pod":/biz_init.sql --namespace=default
kubectl exec "$pod" --namespace=default -- psql -U postgres -f /biz_init.sql

# Europe East
echo "Initializing Europe East - MySQL"
cmd="kubectl exec -i deployments/europe-east-mysql-db --namespace=default -- mysql --user=root --password=password biz <<EOF
SELECT 'READY';
EOF"
wait_for_ready
pod=$(kubectl --namespace=default get pods -o jsonpath="{.items[0].metadata.name}" --selector=app=europe-east-mysql-db)
kubectl cp "$DIR"/mysql.sql "$pod":/biz-init.sql --namespace=default
kubectl exec "$pod" --namespace=default -- sh -c "mysql --user=root --password=password biz < /biz-init.sql"

# Europe West
echo "Initializing Europe West - SQL Server"
cmd="kubectl exec deployments/europe-west-sql-server-db --namespace=default -- /opt/mssql-tools/bin/sqlcmd -U sa -P Str0ngPwd! -Q \"SELECT 'READY'\""
wait_for_ready
pod=$(kubectl --namespace=default get pods -o jsonpath="{.items[0].metadata.name}" --selector=app=europe-west-sql-server-db)
kubectl cp "$DIR"/sqlserver.sql "$pod":/biz-init.sql --namespace=default
kubectl exec "$pod" --namespace=default -- /opt/mssql-tools/bin/sqlcmd -U sa -P Str0ngPwd! -i biz-init.sql

# Asia
echo "Initializing Asia - Oracle"
cmd="kubectl exec -i deployments/asia-oracle-db --namespace=default -- sqlplus sys/manager@XE as sysdba <<EOF
select 'READY' from v\$instance WHERE STATUS = 'OPEN';
EOF"
wait_for_ready
pod=$(kubectl --namespace=default get pods -o jsonpath="{.items[0].metadata.name}" --selector=app=asia-oracle-db)
kubectl cp "$DIR"/oracle.sql "$pod":/biz-init.sql --namespace=default
kubectl exec "$pod" --namespace=default -- sqlplus sys/manager@XE as sysdba @/biz-init.sql
