WHENEVER SQLERROR EXIT 1;

DECLARE
    USEREXIST INTEGER;
BEGIN
    SELECT COUNT(*) INTO USEREXIST FROM ALL_USERS WHERE USERNAME='C##BIZ';
    IF (USEREXIST > 0) THEN
        EXECUTE IMMEDIATE 'DROP USER C##BIZ CASCADE';
    END IF;
END;
/

CREATE USER C##BIZ IDENTIFIED BY "TEST";
GRANT UNLIMITED TABLESPACE TO C##BIZ;
GRANT ALL PRIVILEGES TO C##BIZ;
ALTER SESSION SET current_schema = C##BIZ;

--
-- customer
--

CREATE TABLE customer (
  id number(10) primary key not null,
  customer_id varchar2(25) not null,
  first_name varchar2(100),
  last_name varchar2(100),
  email varchar2(255),
  birth_date varchar(255),
  phone varchar2(25),
  address varchar2(255),
  city varchar2(255),
  state varchar2(2),
  zip varchar2(10),
  sales_rep_id varchar2(25),
  created timestamp with time zone not null,
  updated timestamp with time zone not null
);

CREATE UNIQUE INDEX CUSTOMER_UNIQUE_IDX ON CUSTOMER (CUSTOMER_ID);
CREATE INDEX CUSTOMER_UPDATED_IDX ON CUSTOMER (UPDATED);

CREATE SEQUENCE CUSTOMER_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER CUSTOMER_INSERT_TRIGGER
BEFORE INSERT ON CUSTOMER
FOR EACH ROW
BEGIN
    SELECT CUSTOMER_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER CUSTOMER_UPDATE_TRIGGER
BEFORE UPDATE ON CUSTOMER
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/


--
-- employee
--

CREATE TABLE employee (
  id number(10) primary key not null,
  employee_id varchar2(25) not null,
  first_name varchar2(100),
  last_name varchar2(100),
  email varchar2(255),
  birth_date date,
  phone varchar2(25),
  address varchar2(255),
  city varchar2(255),
  state varchar2(2),
  zip varchar2(10),
  hire_date date,
  primary_department_id varchar2(25),
  manager_employee_id varchar2(25),
  classification_code varchar2(25),
  created timestamp with time zone not null,
  updated timestamp with time zone not null
);

CREATE UNIQUE INDEX EMPLOYEE_UNIQUE_IDX ON EMPLOYEE (EMPLOYEE_ID);
CREATE INDEX EMPLOYEE_UPDATED_IDX ON EMPLOYEE (UPDATED);

CREATE SEQUENCE EMPLOYEE_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER EMPLOYEE_INSERT_TRIGGER
BEFORE INSERT ON EMPLOYEE
FOR EACH ROW
BEGIN
    SELECT EMPLOYEE_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER EMPLOYEE_UPDATE_TRIGGER
BEFORE UPDATE ON EMPLOYEE
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/


--
-- department
--

CREATE TABLE department (
  id number(10) primary key not null,
  department_id varchar2(25) not null,
  name varchar2(100) not null,
  description varchar2(4000),
  created timestamp with time zone not null,
  updated timestamp with time zone not null
);

CREATE UNIQUE INDEX DEPARTMENT_UNIQUE_IDX ON DEPARTMENT (DEPARTMENT_ID);
CREATE INDEX DEPARTMENT_UPDATED_IDX ON DEPARTMENT (UPDATED);

CREATE SEQUENCE DEPARTMENT_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER DEPARTMENT_INSERT_TRIGGER
BEFORE INSERT ON DEPARTMENT
FOR EACH ROW
BEGIN
    SELECT DEPARTMENT_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER DEPARTMENT_UPDATE_TRIGGER
BEFORE UPDATE ON DEPARTMENT
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/

--
-- employee department
--

CREATE TABLE employee_department (
  employee_id varchar2(25) not null,
  department_id varchar2(25) not null,
  employee_role varchar2(100) not null,
  created timestamp with time zone not null,
  updated timestamp with time zone not null,
  primary key(employee_id, department_id)
);

CREATE INDEX EMPLOYEE_DEPT_UPDATED_IDX ON EMPLOYEE_DEPARTMENT (UPDATED);

CREATE OR REPLACE TRIGGER EMPLOYEE_DEPT_INSERT_TRIGGER
BEFORE INSERT ON EMPLOYEE_DEPARTMENT
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER EMPLOYEE_DEPT_UPDATE_TRIGGER
BEFORE UPDATE ON EMPLOYEE_DEPARTMENT
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/

--
-- products (change table method)
--

CREATE TABLE product (
  id number(10) primary key not null,
  product_id varchar2(25) not null,
  name varchar2(100) not null,
  description varchar2(4000),
  price decimal(10,2),
  department_id varchar2(25),
  manufacturer_id varchar2(25),
  created timestamp with time zone not null
);

CREATE TABLE product_change_history (
  product_change_history_id number(10) primary key not null,
  id number(10) not null,
  product_id varchar2(25) not null,
  change_type varchar2(20) not null,
  change_time timestamp with time zone not null
);


CREATE UNIQUE INDEX PRODUCT_UNIQUE_IDX ON PRODUCT (PRODUCT_ID);

CREATE SEQUENCE PRODUCT_SEQ START WITH 1;
CREATE SEQUENCE PRODUCT_HISTORY_SEQ START WITH 1;

CREATE OR REPLACE TRIGGER PRODUCT_INSERT_TRIGGER
BEFORE INSERT ON PRODUCT
FOR EACH ROW
BEGIN
    SELECT PRODUCT_SEQ.nextval, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER PRODUCT_UPDATE_TRIGGER
AFTER INSERT OR UPDATE OR DELETE ON PRODUCT
FOR EACH ROW
BEGIN
    IF INSERTING  THEN
        INSERT INTO product_change_history (product_change_history_id, id, product_id, change_time, change_type)
        SELECT PRODUCT_HISTORY_SEQ.nextval, :NEW.id, :NEW.product_id,  CURRENT_TIMESTAMP, 'INSERT' FROM DUAL;
    ELSIF UPDATING THEN
        INSERT INTO product_change_history (product_change_history_id, id, product_id, change_time, change_type)
        SELECT PRODUCT_HISTORY_SEQ.nextval, :NEW.id, :NEW.product_id,  CURRENT_TIMESTAMP, 'UPDATE'  FROM DUAL;
    ELSE
        INSERT INTO product_change_history (product_change_history_id, id, product_id, change_time, change_type)
        SELECT PRODUCT_HISTORY_SEQ.nextval, :OLD.id, :OLD.product_id,  CURRENT_TIMESTAMP, 'DELETE'  FROM DUAL;
    END IF;
END;
/

--
-- purchase_order
--

CREATE TABLE purchase_order (
   id number(10) primary key not null,
   purchase_order_id varchar2(25) not null,
   customer_id varchar2(25) not null,
   employee_id varchar2(25) not null,
   promotion_id varchar2(25),
   description varchar2(4000),
   total decimal(10,2),
   created timestamp with time zone not null,
   updated timestamp with time zone not null
);

CREATE UNIQUE INDEX PURCHASE_ORDER_UNIQUE_IDX ON PURCHASE_ORDER (PURCHASE_ORDER_ID);
CREATE INDEX PURCHASE_ORDER_UPDATED_IDX ON PURCHASE_ORDER (UPDATED);

CREATE SEQUENCE PURCHASE_ORDER_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER PURCHASE_ORDER_INSERT_TRIGGER
BEFORE INSERT ON PURCHASE_ORDER
FOR EACH ROW
BEGIN
    SELECT PURCHASE_ORDER_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER PURCHASE_ORDER_UPDATE_TRIGGER
BEFORE UPDATE ON PURCHASE_ORDER
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/

--
-- sale
--

CREATE TABLE sale (
      id number(10) primary key not null,
      sale_id varchar2(25) not null,
      purchase_order_id varchar2(25) not null,
      product_id varchar2(25) not null,
      quantity number(38) not null,
      unit_price decimal(10,2) not null,
      discount decimal(10,2),
      total decimal(10,2) not null,
      description varchar2(4000),
      created timestamp with time zone not null,
      updated timestamp with time zone not null
);


CREATE UNIQUE INDEX SALE_UNIQUE_IDX ON SALE (SALE_ID);
CREATE INDEX SALE_UPDATED_IDX ON SALE (UPDATED);

CREATE SEQUENCE SALE_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER SALE_INSERT_TRIGGER
BEFORE INSERT ON SALE
FOR EACH ROW
BEGIN
    SELECT SALE_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER SALE_UPDATE_TRIGGER
BEFORE UPDATE ON SALE
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/

--
-- promotion
--

CREATE TABLE promotion (
      id number(10) primary key not null,
      promotion_id varchar2(25) not null,
      description varchar2(4000) not null,
      discount decimal(10,2) not null,
      created timestamp with time zone not null,
      updated timestamp with time zone not null
);


CREATE UNIQUE INDEX PROMOTION_UNIQUE_IDX ON PROMOTION (PROMOTION_ID);
CREATE INDEX PROMOTION_UPDATED_IDX ON PROMOTION (UPDATED);

CREATE SEQUENCE PROMOTION_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER PROMOTION_INSERT_TRIGGER
BEFORE INSERT ON PROMOTION
FOR EACH ROW
BEGIN
    SELECT PROMOTION_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER PROMOTION_UPDATE_TRIGGER
BEFORE UPDATE ON PROMOTION
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/


--
-- manufacturer
--

CREATE TABLE manufacturer (
      id number(10) primary key not null,
      manufacturer_id varchar2(25) not null,
      name varchar2(255) not null,
      location varchar2(4000),
      created timestamp with time zone not null,
      updated timestamp with time zone not null
);


CREATE UNIQUE INDEX MANUFACTURER_UNIQUE_IDX ON MANUFACTURER (MANUFACTURER_ID);
CREATE INDEX MANUFACTURER_UPDATED_IDX ON MANUFACTURER (UPDATED);

CREATE SEQUENCE MANUFACTURER_SEQ START WITH 1;
CREATE OR REPLACE TRIGGER MANUFACTURER_INSERT_TRIGGER
BEFORE INSERT ON MANUFACTURER
FOR EACH ROW
BEGIN
    SELECT MANUFACTURER_SEQ.nextval, CURRENT_TIMESTAMP, CURRENT_TIMESTAMP
    INTO :NEW.id, :NEW.created, :NEW.updated
    FROM dual;
END;
/

CREATE OR REPLACE TRIGGER MANUFACTURER_UPDATE_TRIGGER
BEFORE UPDATE ON MANUFACTURER
FOR EACH ROW
BEGIN
    SELECT CURRENT_TIMESTAMP
    INTO :NEW.updated
    FROM dual;
END;
/