\set ON_ERROR_STOP on

DROP DATABASE IF EXISTS biz;
CREATE DATABASE biz;

\c biz;

--
-- generic timestamp trigger
--

CREATE OR REPLACE FUNCTION public.update_updated_column()
RETURNS trigger
LANGUAGE plpgsql
AS $$
BEGIN
    NEW.updated = NOW();
    RETURN NEW;
END;
$$;

--
-- customer
--

CREATE TABLE public.customer (
  id bigserial primary key not null,
  customer_id varchar(25) unique not null,
  first_name varchar(100),
  last_name varchar(100),
  email varchar(255),
  birth_date varchar(255),
  phone varchar(25),
  address varchar(255),
  city varchar(255),
  state varchar(2),
  zip varchar(10),
  sales_rep_id varchar(25),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX customer_updated_idx ON public.customer (updated);
CREATE TRIGGER customer_updated BEFORE UPDATE OR INSERT ON public.customer
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();

--
-- employee
--

CREATE TABLE public.employee (
  id bigserial primary key not null,
  employee_id varchar(25) unique not null,
  first_name varchar(100),
  last_name varchar(100),
  email varchar(255),
  birth_date date,
  phone varchar(25),
  address varchar(255),
  city varchar(255),
  state varchar(2),
  zip varchar(10),
  hire_date date,
  primary_department_id varchar(25),
  manager_employee_id varchar(25),
  classification_code varchar(25),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX employee_updated_idx ON public.employee (updated);
CREATE TRIGGER employee_updated BEFORE UPDATE OR INSERT ON public.employee
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();


--
-- department
--

CREATE TABLE public.department (
  id bigserial primary key not null,
  department_id varchar(25) unique not null,
  name varchar(100) not null,
  description varchar(4000),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX department_updated_idx ON public.department (updated);
CREATE TRIGGER department_updated BEFORE UPDATE OR INSERT ON public.department
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();


--
-- employee department
--

CREATE TABLE public.employee_department (
  employee_id varchar(25) not null,
  department_id varchar(25) not null,
  employee_role varchar(100) not null,
  created timestamptz not null default now(),
  updated timestamptz not null default now(),
  primary key(employee_id, department_id)
);

CREATE INDEX employee_department_updated_idx ON public.employee_department (updated);
CREATE TRIGGER employee_department_updated BEFORE UPDATE OR INSERT ON public.employee_department
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();


--
-- products (change table method)
--

CREATE TABLE public.product (
  id bigserial primary key not null,
  product_id varchar(25) unique not null,
  name varchar(100) not null,
  description varchar(4000),
  price decimal(10,2),
  department_id varchar(25),
  manufacturer_id varchar(25),
  created timestamptz not null default now()
);

CREATE TABLE public.product_change_history (
  product_change_history_id bigserial primary key not null,
  id bigint not null,
  product_id varchar(25) not null,
  change_type varchar(20) not null,
  change_time timestamptz not null default now()
);

CREATE OR REPLACE FUNCTION public.update_product_change_history()
RETURNS trigger
LANGUAGE plpgsql
AS $$
BEGIN
    INSERT INTO public.product_change_history (id, product_id, change_type)
    VALUES(CASE WHEN TG_OP = 'DELETE' THEN OLD.id ELSE NEW.id END,
           CASE WHEN TG_OP = 'DELETE' THEN OLD.product_id ELSE NEW.product_id END,
           TG_OP);
    RETURN NEW;
END;
$$;

CREATE TRIGGER product_changed AFTER UPDATE OR INSERT OR DELETE ON public.product
FOR EACH ROW EXECUTE PROCEDURE update_product_change_history();

--
-- purchase_order
--

CREATE TABLE public.purchase_order (
  id bigserial primary key not null,
  purchase_order_id varchar(25) unique not null,
  customer_id varchar(25) not null,
  employee_id varchar(25) not null,
  promotion_id varchar(25),
  description varchar(4000),
  total decimal(10,2),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX purchase_order_updated_idx ON public.purchase_order (updated);
CREATE TRIGGER purchase_order_updated BEFORE UPDATE OR INSERT ON public.purchase_order
    FOR EACH ROW EXECUTE PROCEDURE update_updated_column();

--
-- sale
--

CREATE TABLE public.sale (
  id bigserial primary key not null,
  sale_id varchar(25) unique not null,
  purchase_order_id varchar(25) not null,
  product_id varchar(25) not null,
  quantity integer not null,
  unit_price decimal(10,2) not null,
  discount decimal(10,2),
  total decimal(10,2) not null,
  description varchar(4000),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX sale_updated_idx ON public.sale (updated);
CREATE TRIGGER sale_updated BEFORE UPDATE OR INSERT ON public.sale
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();

--
-- promotion
--

CREATE TABLE public.promotion (
  id bigserial primary key not null,
  promotion_id varchar(25) unique not null,
  description varchar(4000) not null,
  discount decimal(10,2) not null,
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX promotion_updated_idx ON public.promotion (updated);
CREATE TRIGGER promotion_updated BEFORE UPDATE OR INSERT ON public.promotion
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();


--
-- manufacturer
--

CREATE TABLE public.manufacturer (
  id bigserial primary key not null,
  manufacturer_id varchar(25) unique not null,
  name varchar(255) not null,
  location varchar(4000),
  created timestamptz not null default now(),
  updated timestamptz not null default now()
);

CREATE INDEX manufacturer_updated_idx ON public.manufacturer (updated);
CREATE TRIGGER manufacturer_updated BEFORE UPDATE OR INSERT ON public.manufacturer
FOR EACH ROW EXECUTE PROCEDURE update_updated_column();
