DROP DATABASE IF EXISTS biz;
GO
CREATE DATABASE biz;
GO

USE biz;
GO

--
-- customer
--

CREATE TABLE customer (
  id bigint identity primary key not null,
  customer_id varchar(25) unique not null,
  first_name varchar(100),
  last_name varchar(100),
  email varchar(255),
  birth_date varchar(255),
  phone varchar(25),
  address varchar(255),
  city varchar(255),
  state varchar(2),
  zip varchar(10),
  sales_rep_id varchar(25),
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX customer_updated_idx ON customer (updated);
GO

CREATE TRIGGER customer_inserted ON customer AFTER INSERT AS
UPDATE customer SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN customer c ON i.id = c.id;
GO

CREATE TRIGGER customer_updated ON customer AFTER UPDATE AS
UPDATE customer SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN customer c ON i.id = c.id;
GO

--
-- employee
--

CREATE TABLE employee (
  id bigint identity primary key not null,
  employee_id varchar(25) unique not null,
  first_name varchar(100),
  last_name varchar(100),
  email varchar(255),
  birth_date date,
  phone varchar(25),
  address varchar(255),
  city varchar(255),
  state varchar(2),
  zip varchar(10),
  hire_date date,
  primary_department_id varchar(25),
  manager_employee_id varchar(25),
  classification_code varchar(25),
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX employee_updated_idx ON employee (updated);
GO

CREATE TRIGGER employee_inserted ON employee AFTER INSERT AS
UPDATE employee SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN employee c ON i.id = c.id;
GO

CREATE TRIGGER employee_updated ON employee AFTER UPDATE AS
UPDATE employee SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN employee c ON i.id = c.id;
GO

--
-- department
--

CREATE TABLE department (
  id bigint identity primary key not null,
  department_id varchar(25) unique not null,
  name varchar(100) not null,
  description varchar(4000),
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX department_updated_idx ON department (updated);
GO

CREATE TRIGGER department_inserted ON department AFTER INSERT AS
UPDATE department SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN department c ON i.id = c.id;
GO

CREATE TRIGGER department_updated ON department AFTER UPDATE AS
UPDATE department SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN department c ON i.id = c.id;
GO

--
-- employee department
--

CREATE TABLE employee_department (
  employee_id varchar(25) not null,
  department_id varchar(25) not null,
  employee_role varchar(100) not null,
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP,
  primary key(employee_id, department_id)
);

CREATE INDEX employee_department_updated_idx ON employee_department (updated);
GO

CREATE TRIGGER employee_department_inserted ON employee_department AFTER INSERT AS
UPDATE employee_department SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN employee_department c ON i.employee_id = c.employee_id
   AND i.department_id = c.department_id;
GO

CREATE TRIGGER employee_department_updated ON employee_department AFTER UPDATE AS
UPDATE employee_department SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN employee_department c ON i.employee_id = c.employee_id
   AND i.department_id = c.department_id;
GO

--
-- products (change table method)
--

CREATE TABLE product (
  id bigint identity primary key not null,
  product_id varchar(25) unique not null,
  name varchar(100) not null,
  description varchar(4000),
  price decimal(10,2),
  department_id varchar(25),
  manufacturer_id varchar(25),
  created datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE TABLE product_change_history (
  product_change_history_id bigint identity primary key not null,
  id bigint not null,
  product_id varchar(25) not null,
  change_type varchar(20) not null,
  change_time datetime2(6) not null default CURRENT_TIMESTAMP
);

GO

CREATE TRIGGER product_inserted ON product AFTER INSERT AS
INSERT INTO product_change_history (id, product_id, change_type, change_time)
SELECT i.id, i.product_id, 'INSERT', CURRENT_TIMESTAMP
  FROM Inserted i;
GO

CREATE TRIGGER product_updated ON product AFTER UPDATE AS
INSERT INTO product_change_history (id, product_id, change_type, change_time)
SELECT i.id, i.product_id, 'UPDATE', CURRENT_TIMESTAMP
  FROM Inserted i;
GO

CREATE TRIGGER product_deleted ON product AFTER DELETE AS
INSERT INTO product_change_history (id, product_id, change_type, change_time)
SELECT d.id, d.product_id, 'DELETE', CURRENT_TIMESTAMP
  FROM Deleted d;
GO

--
-- purchase order
--

CREATE TABLE purchase_order (
   id bigint identity primary key not null,
   purchase_order_id varchar(25) unique not null,
   customer_id varchar(25) not null,
   employee_id varchar(25) not null,
   promotion_id varchar(25),
   description varchar(4000),
   total decimal(10,2),
   created datetime2(6) not null default CURRENT_TIMESTAMP,
   updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX purchase_order_updated_idx ON purchase_order (updated);
GO

CREATE TRIGGER purchase_order_inserted ON purchase_order AFTER INSERT AS
UPDATE purchase_order SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN purchase_order c ON i.id = c.id;
GO

CREATE TRIGGER purchase_order_updated ON purchase_order AFTER UPDATE AS
UPDATE purchase_order SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN purchase_order c ON i.id = c.id;
GO

--
-- sale
--

CREATE TABLE sale (
  id bigint identity primary key not null,
  sale_id varchar(25) unique not null,
  purchase_order_id varchar(25) not null,
  product_id varchar(25) not null,
  quantity int not null,
  unit_price decimal(10,2) not null,
  discount decimal(10,2),
  total decimal(10,2) not null,
  description varchar(4000) not null,
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX sale_updated_idx ON sale (updated);
GO

CREATE TRIGGER sale_inserted ON sale AFTER INSERT AS
UPDATE sale SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN sale c ON i.id = c.id;
GO

CREATE TRIGGER sale_updated ON sale AFTER UPDATE AS
UPDATE sale SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN sale c ON i.id = c.id;
GO

--
-- promotion
--

CREATE TABLE promotion (
  id bigint identity primary key not null,
  promotion_id varchar(25) unique not null,
  description varchar(4000) not null,
  discount decimal(10,2) not null,
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX promotion_updated_idx ON promotion (updated);
GO

CREATE TRIGGER promotion_inserted ON promotion AFTER INSERT AS
UPDATE promotion SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN promotion c ON i.id = c.id;
GO

CREATE TRIGGER promotion_updated ON promotion AFTER UPDATE AS
UPDATE promotion SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN promotion c ON i.id = c.id;
GO


--
-- manufacturer
--

CREATE TABLE manufacturer (
  id bigint identity primary key not null,
  manufacturer_id varchar(25) unique not null,
  name varchar(255) not null,
  location varchar(4000),
  created datetime2(6) not null default CURRENT_TIMESTAMP,
  updated datetime2(6) not null default CURRENT_TIMESTAMP
);

CREATE INDEX manufacturer_updated_idx ON manufacturer (updated);
GO

CREATE TRIGGER manufacturer_inserted ON manufacturer AFTER INSERT AS
UPDATE manufacturer SET created = CURRENT_TIMESTAMP, updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN manufacturer c ON i.id = c.id;
GO

CREATE TRIGGER manufacturer_updated ON manufacturer AFTER UPDATE AS
UPDATE manufacturer SET updated = CURRENT_TIMESTAMP
  FROM Inserted i INNER JOIN manufacturer c ON i.id = c.id;
GO
