import json
import os
import subprocess
import sys
from string import Template

from functions import do_get
from functions import do_patch
from functions import exec_command
from functions import get_token
from functions import get_zone
from functions import get_zone_adaptor_registration
from functions import k8s_apply_stdin
from functions import k8s_drain
from functions import load_global_file
from functions import post_adaptor

assert sys.version_info >= (3, 7)

#
# Configure adaptors.
#
# First, adaptors are registered with YOUnite and their credentials are stored in
# a list of dictionaries. Finally, th dictionaries are read and adaptors are created
# with those credentials, along with the specific configuration information from
# each.

script_path = os.path.dirname(os.path.realpath(__file__))
adaptors_path = script_path + "/../adaptors/"
adaptor_file = script_path + "/../adaptors/kubernetes/adaptor.yml"
template_file = script_path + "/../adaptors/adaptor_template.json"

# globals
log_level = "INFO"
concurrency = "4"
mongo_url = "mongodb://admin:admin@mongo/admin"

tables_files = []
for x in os.listdir(script_path + "/../adaptors/tables"):
    if x.endswith(".json"):
        tables_files.append(script_path + "/../adaptors/tables/" + x)


def create_adaptor(adaptor_file, global_file, metadata_file, tables_files, adaptor_type_uuid):
    """
    Create an adaptor and return its registration information. If the adaptor already exists, the
    existing adaptor's registration will be returned. If the zone does not exist, NotFoundError
    will be thrown.
    """

    af = open(adaptor_file, "r")

    try:
        adaptor_file_json = json.load(af)

        f = open(metadata_file, "r")
        try:
            metadata_json = json.load(f)
            if tables_files is not None:
                metadata_json["metadata"]["tables"] = []
                for fn in tables_files:
                    tf = open(fn, "r")
                    try:
                        metadata_json["metadata"]["tables"].append(json.load(tf))
                    finally:
                        tf.close()

            # Convert the metadata to a string to perform replacement, then convert back
            metadata = json.dumps(metadata_json["metadata"])
            if "jdbc_url" in adaptor_file_json:
                metadata = metadata.replace("{{JDBC_URL}}", adaptor_file_json['jdbc_url'])
            if "container_name" in adaptor_file_json:
                metadata = metadata.replace("{{CONTAINER_NAME}}", adaptor_file_json['container_name'])
            metadata = json.loads(metadata)
        finally:
            f.close()

        print("*** ADAPTOR INFORMATION ***")
        print("Name:         " + adaptor_file_json['adaptor_name'] + "(container: " + adaptor_file_json['container_name'] + ")")
        print("Zone:         " + adaptor_file_json['adaptor_zone'])

        f = load_global_file(global_file)
        dgs_token = get_token(f["token_href"], f["client_id"], f["client_secret"], f["dgs_username"], f["dgs_password"])

        adaptor_uuid = post_adaptor(f["href"], dgs_token, adaptor_file_json, metadata_json["capabilities"], metadata, 
                                    adaptor_type_uuid)

        # GET registration
        zone_uuid = get_zone(f["href"], dgs_token, adaptor_file_json['adaptor_zone'])
        registration = get_zone_adaptor_registration(f["href"], dgs_token, zone_uuid, adaptor_uuid)
    finally:
        af.close()

    return registration


def __patch_adaptor_state(global_file: str, zone_name: str, adaptor_uuid: str, state: str):
    f = load_global_file(global_file)
    token = get_token(f["token_href"], f["client_id"], f["client_secret"], f["dgs_username"], f["dgs_password"])

    zone_uuid = get_zone(f["href"], token, zone_name)

    adaptor = do_get(f["href"] + "/zones/" + zone_uuid + "/adaptors/" + adaptor_uuid, token)

    print("Updating adaptor state: " + state)

    do_patch(f["href"] + '/zones/' + zone_uuid + '/adaptors/' + adaptor_uuid, token, json.dumps({
        "changeVersion": adaptor["changeVersion"],
        "state": state
    }))


def pause_adaptor(global_file: str, zone_name: str, adaptor_uuid: str):
    """Instruct an adaptor to pause. If the adaptor is not found, NotFoundError is thrown."""
    __patch_adaptor_state(global_file, zone_name, adaptor_uuid, "PAUSE")


def start_adaptor(global_file: str, zone_name: str, adaptor_uuid: str):
    """Instruct an adaptor to start. If the adaptor is not found, NotFoundError is thrown."""
    __patch_adaptor_state(global_file, zone_name, adaptor_uuid, "PLAY")


def create_adaptors(global_file, adaptor_number=None):
    """Create adaptors for the biz demo. Passing a value to adaptor_number will only create that one adaptor."""

    adaptor_dict = {}

    # adaptor type of the "db adaptor"
    adaptor_type_uuid = "f5e35921-e3bc-4e28-bbd8-75fa596a6b9a"

    # Asia adaptor -
    if adaptor_number is None or adaptor_number == "1":
        print("Creating Asia Adaptor")
        asia_adaptor = create_adaptor(adaptors_path + "asia-adaptor.json", global_file, template_file, tables_files, adaptor_type_uuid)
        if asia_adaptor:
            adaptor_dict["asia-adaptor"] = asia_adaptor

    # Europe West adaptor -
    if adaptor_number is None or adaptor_number == "2":
        print("Creating Europe West Adaptor")
        europe_west_adaptor = create_adaptor(adaptors_path + "europe-west-adaptor.json", global_file, template_file, tables_files, adaptor_type_uuid)
        if europe_west_adaptor:
            adaptor_dict["europe-west-adaptor"] = europe_west_adaptor

    # North America adaptor -
    if adaptor_number is None or adaptor_number == "3":
        print("Creating North America Adaptor")
        north_america_adaptor = create_adaptor(adaptors_path + "north-america-adaptor.json", global_file, template_file, tables_files, adaptor_type_uuid)
        if north_america_adaptor:
            adaptor_dict["north-america-adaptor"] = north_america_adaptor

    # Europe East adaptor -
    if adaptor_number is None or adaptor_number == "4":
        print("Creating Europe East Adaptor")
        europe_east_adaptor = create_adaptor(adaptors_path + "europe-east-adaptor.json", global_file, template_file, tables_files, adaptor_type_uuid)
        if europe_east_adaptor:
            adaptor_dict["europe-east-adaptor"] = europe_east_adaptor

    return adaptor_dict


def start_adaptor_local(db_adaptor_image, c, container_name, java_opts, message_bus_url, environment_vars=None):
    """Start one locally running adaptor"""

    cmd = ["docker", "run", "--name", container_name, "--network", "younite", "--expose", "8001",
           "-e", "ADAPTOR_UUID=" + c["uuid"],
           "-e", "CLIENT_ID=" + c["clientId"],
           "-e", "CLIENT_SECRET=" + c["clientSecret"],
           "-e", "MESSAGE_BUS_URL=" + message_bus_url,
           "-e", "MONGO_URL=" + mongo_url,
           "-e", "LOG_LEVEL=" + log_level,
           "-e", "CONCURRENCY=" + str(concurrency),
           "-e", "JAVA_OPTS=" + java_opts]

    if environment_vars is not None:
        for k in environment_vars:
            cmd += ["-e", k + "=" + str(environment_vars[k])]

    cmd += ["-d", db_adaptor_image]

    exec_command(cmd)
    print("")


def start_adaptors_local(db_adaptor_image, message_bus_url, adaptor_creds):
    """Start all biz demo adaptors locally. Only those with credentials in adaptor_creds will be started."""

    if "asia-adaptor" in adaptor_creds:
        start_adaptor_local(db_adaptor_image, adaptor_creds["asia-adaptor"], "asia-adaptor", "", message_bus_url,
                            {"JDBC_USER": "postgres", "JDBC_PASSWORD": "mysecretpassword"})
    if "europe-west-adaptor" in adaptor_creds:
        start_adaptor_local(db_adaptor_image, adaptor_creds["europe-west-adaptor"], "europe-west-adaptor", "", message_bus_url,
                              {"JDBC_USER": "sa", "JDBC_PASSWORD": "Str0ngPwd!"})
    if "north-america-adaptor" in adaptor_creds:
        start_adaptor_local(db_adaptor_image, adaptor_creds["north-america-adaptor"], "north-america-adaptor", "", message_bus_url,
                              {"JDBC_USER": "postgres", "JDBC_PASSWORD": "mysecretpassword"})
    if "europe-east-adaptor" in adaptor_creds:
        start_adaptor_local(db_adaptor_image, adaptor_creds["europe-east-adaptor"], "europe-east-adaptor", "", message_bus_url,
                              {"JDBC_USER": "biz", "JDBC_PASSWORD": "password"})


def stop_adaptors_local(db_adaptor_image):
    """Stop all running adaptors with the given image name"""

    running = subprocess.check_output(["docker", "ps", "-q", "--filter", "ancestor=" + db_adaptor_image], text=True)
    if len(running) > 0:
        for c in running.split("\n"):
            if len(c) > 0:
                exec_command(["docker", "stop", c])
    stopped = subprocess.check_output(["docker", "ps", "-a", "-q", "--filter", "ancestor=" + db_adaptor_image], text=True)
    if len(stopped) > 0:
        for c in stopped.split("\n"):
            if len(c) > 0:
                exec_command(["docker", "rm", c])


def __start_adaptor_aws_demo(namespace, template, c, container_name, java_opts, message_bus_url, jdbc_user, jdbc_password):
    """Start an adaptor in the AWS demo environment"""

    adaptor_uuid = c["uuid"]
    client_id = c["clientId"]
    client_secret = c["clientSecret"]
    image_tag = "latest"
    git_commit = ""

    yml = Template(Template(template).safe_substitute(locals())).safe_substitute(globals())
    k8s_apply_stdin(yml, namespace)


def start_adaptors_aws_demo(namespace, adaptor_creds):
    """Start all biz demo adaptors in the AWS demo environment"""

    message_bus_url = "tcp://younite-mb:61616"
    f = open(adaptor_file)
    try:
        template = f.read()
        __start_adaptor_aws_demo(namespace, template, adaptor_creds["asia-adaptor"], "asia-adaptor", "-Doracle.jdbc.timezoneAsRegion=false", message_bus_url, "C##BIZ", "TEST")
        __start_adaptor_aws_demo(namespace, template, adaptor_creds["north-america-adaptor"], "north-america-adaptor", "", message_bus_url, "postgres", "mysecretpassword")
        __start_adaptor_aws_demo(namespace, template, adaptor_creds["europe-west-adaptor"], "europe-west-adaptor", "", message_bus_url, "sa", "Str0ngPwd!")
        __start_adaptor_aws_demo(namespace, template, adaptor_creds["europe-east-adaptor"], "europe-east-adaptor", "", message_bus_url, "biz", "password")
    finally:
        f.close()


def stop_adaptors_aws_demo(namespace):
    """Stop all biz demo adaptors in the AWS demo environment"""

    apps = ["north-america-adaptor", "europe-west-adaptor", "europe-east-adaptor", "asia-adaptor"]
    k8s_drain(apps, namespace)
