import json
import subprocess
import sys
import time
import os
import requests

assert sys.version_info >= (3, 7)


class YouniteError(Exception):
    """Base class for exceptions in this module."""
    pass


class RequestError(YouniteError):
    """
    Exception arising from a REST request

    Attributes:
        status_code -- status code
        message -- explanation of the error
    """

    def __init__(self, status_code, message):
        self.status_code = status_code
        self.message = message


class GeneralError(YouniteError):
    """
    General error with a message

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class NotFoundError(YouniteError):
    """
    Error arising when an entity cannot be found by name

    Attributes:
        entity -- entity type
        name -- name of the entity
    """

    def __init__(self, entity, name):
        self.entity = entity
        self.name = name


def wait_for_ready(cmd: list, stdin: str = None, max_attempts: int = 30, sleep_time: int = 5):
    """
    Wait for a process to be "ready". The command is executed and if the
    output contains the word "READY", then the process is ready, otherwise
    it is retried until the process becomes ready or the max attempts is
    reached.
    """
    attempts = 1
    if not type(cmd) is list:
        cmd = [cmd]

    while True:
        print("executing: " + " ".join(cmd))
        result = subprocess.run(args=cmd, input=stdin, capture_output=True, text=True)
        print(result.stdout)
        if result.returncode == 0:
            if "READY" in result.stdout:
                break

        attempts = attempts + 1

        if attempts > max_attempts:
            print("Max wait time exceeded waiting for database to start")
            sys.exit(1)

        print("Call failed ... attempting to retry in 5 seconds ...")
        time.sleep(sleep_time)


def exec_command(cmd, stdin=None, shell=False):
    """Execute a program / shell command with error handling. GeneralError is thrown if the command fails."""
    if type(cmd) is list:
        print("executing: " + " ".join(cmd))
    else:
        print("executing: " + cmd)
    result = subprocess.run(args=cmd, input=stdin, text=True, shell=shell)
    if result.returncode != 0:
        raise GeneralError("ERROR! Command failed with exit code " + str(result.returncode))


def replace_hostname (uri: str) -> str:
    """
    If the environment variable YOUNITE_HOSTNAME is set and not null, this
    replaces the default hostname "local.younite.org" with the value of YOUNITE_HOSTNAME
    """
    younite_hostname = os.environ.get("YOUNITE_HOSTNAME")

    if younite_hostname is None or younite_hostname.strip() == "":
        return uri

    if "local.younite.org" not in uri:
        return uri

    return uri.replace("local.younite.org", younite_hostname)

def list_to_dict(pairs: list) -> dict:
    """
    Convert a list of key/value pairs to a dictionary.
    This is useful for processing environment variables
    which come as lists of objects with a "key" and a "value".
    """

    out = {}
    for p in pairs:
        out[p["key"]] = p["value"]
    return out


def k8s_get_pod_name(app, namespace):
    """Get the name of a pod in Kubernetes based on the namespace and app"""

    cmd = ["kubectl", "--namespace=" + namespace, "get", "pods", "-o",
           "jsonpath=\"{.items[*].metadata.name}\"", "--selector=app=" + app]
    print("executing " + " ".join(cmd))
    out = subprocess.check_output(args=cmd, text=True).replace("\"", "")
    if len(out) > 0:
        return out.splitlines()[0]
    return None


def k8s_apply(file_name, namespace):
    """Apply a kubernetes file"""
    exec_command(["kubectl", "--namespace=" + namespace, "apply", "-f", file_name])


def k8s_cp(pod, src_file_name, dest_file_name, namespace):
    """Copy a file to a Kubernetes pod"""
    exec_command(["kubectl", "cp", src_file_name, pod + ":" + dest_file_name, "--namespace=" + namespace])


def k8s_exec(pod, cmd, namespace):
    """Execute a command in the shell of a Kubernetes pod"""
    if not type(cmd) is list:
        cmd = [cmd]
    exec_command(["kubectl", "exec", pod, "--namespace=" + namespace, "--"] + cmd)


def k8s_apply_stdin(stdin, namespace):
    """Apply a kubernetes file from standard input. Uses the shell to send the input."""
    exec_command("kubectl --namespace=" + namespace + " apply -f -", stdin, shell=True)


def k8s_drain(apps, namespace):
    """Drain a list of apps in Kubernetes. Scales the deployments to zero, then waits for all pods to be vacated."""

    if type(apps) is not list:
        apps = [apps]

    # scale the apps to 0
    for a in apps:
        k8s_scale(a, 0, namespace)

    # wait for all pods to be vacated
    print("Waiting for pods to be vacated ... ")
    for i in range(60):
        active = False
        
        for a in apps:
            pn = k8s_get_pod_name(a, namespace)
            if pn is not None:
                active = True
                print("Sleeping for 5 seconds ...")
                time.sleep(5)
                break
        if not active:
            break

    if active:
        raise GeneralError("ERROR: Time expired waiting for pods to be vacated")


def k8s_scale(app, replicas, namespace):
    """
    Scale an app to the desired number of replicas. If no deployments exist for the
    app, False is returned. Otherwise the deployment is scaled and True is returned.
    """
    
    cmd = ["kubectl", "--namespace=" + namespace, "get", "deployments",
           "-o", "jsonpath=\"{.items[*].metadata.name}\"", "--selector=app=" + app]

    print("executing " + " ".join(cmd))
    deployment = subprocess.check_output(args=cmd, text=True).replace("\"", "")
    if len(deployment) > 0:
        exec_command(["kubectl", "--namespace=" + namespace, "scale", "--replicas=" + str(replicas),
                      "deployments/" + deployment])
        return True

    return False


def sqlserver_cmd(cmd=None):
    """Construct a SQL Server command"""
    
    base = ["/opt/mssql-tools/bin/sqlcmd", "-U", "sa", "-P", "Str0ngPwd!", "-b", "-V", "1"]
    if cmd is None:
        return base
    if type(cmd) is list:
        return base + cmd
    else:
        return base + [cmd]


def oracle_cmd(cmd=None):
    """Construct an Oracle command"""
    base = ["sqlplus", "sys/manager@XE", "as", "sysdba"]
    if cmd is None:
        return base
    if type(cmd) is list:
        return base + cmd
    else:
        return base + [cmd]


def postgresql_cmd(cmd=None):
    """Construct a PostgreSQL command"""
    
    base = ["psql", "-U", "postgres"]
    if cmd is None:
        return base
    if type(cmd) is list:
        return base + cmd
    else:
        return base + [cmd]


def mysql_cmd(cmd=None):
    """Construct a MySQL command"""
    
    base = ["mysql", "--user=root", "--password=password", "biz"]
    if cmd is None:
        return base
    if type(cmd) is list:
        return base + cmd
    else:
        return base + [cmd]


def read_file(file_name: str) -> str:
    """Read the contents of a file"""
    
    f = open(file_name, "r")
    try:
        return f.read()
    finally:
        f.close()


def load_global_file(global_file: str) -> dict:
    """Load a global file in postman format and convert it to a dictionary of key/value pairs"""

    gf = open(global_file, "r")
    try:
        data = json.load(gf)
        if type(data) is dict:
            values = data["values"]
            if type(values) is list:
                result = {}
                for i in values:
                    if type(i) is dict:
                        key = i["key"]
                        val = i["value"]
                        if type(key) is str and len(key) > 0:
                            result[key] = val
                return result
        raise Exception("Invalid global file, cannot load")
    finally:
        gf.close()


def get_token(uri: str, client_id: str, client_secret: str, username: str, password: str) -> str:
    """Get a token from the Keycloak server"""
    
    req_body = {
        'client_id': client_id,
        'client_secret': client_secret,
        'grant_type': 'password',
        'username': username,
        'password': password,
        'scope': 'openid email'
    }
    response = requests.post(replace_hostname(uri), data=req_body)
    if response.status_code == 200:
        body = response.json()
        token = body.get('access_token')
        if type(token) is str and len(token) > 0:
            return token
        raise RequestError(response.status_code, "Token not found in response body")
    else:
        raise RequestError(response.status_code, "Error retrieving token from Keycloak, status code = "
                           + str(response.status_code))


def __default_headers(auth_token: str) -> dict:
    """The default headers required for requests to the YOUnite API (including bearer token)"""

    return {
        'Authorization': 'Bearer ' + auth_token,
        'Content-Type': 'application/json'
    }


def __handle_response(uri: str, response, action: str) -> dict:
    """
    Handle a REST request response and throw RequestError if the request did not produce a 2xx response.
    If the request returned a response, it will be parsed as JSON, if possible and converted to an object.
    """

    if 200 <= response.status_code < 300:
        try:
            if response.text is not None and len(response.text) > 0:
                return response.json()
        except ValueError:
            # thrown if the response is not JSON - ignore
            pass
    else:
        msg = 'Request failed: ' + action + ' ' + uri
        print(msg)
        print('status code    : ' + str(response.status_code))
        print('response reason: ' + response.reason)
        print('response text  : ' + response.text)
        raise RequestError(response.status_code, msg)

    return {}


def do_post(uri: str, auth_token: str, req_body: str, params=None) -> dict:
    """Perform a POST on an endpoint of the YOUnite API"""

    response = requests.post(replace_hostname(uri), params=params, headers=__default_headers(auth_token), data=req_body)
    return __handle_response(replace_hostname(uri), response, "POST")


def do_put(uri: str, auth_token: str, req_body: str, params=None) -> dict:
    """Perform a PUT on an endpoint of the YOUnite API"""

    response = requests.put(replace_hostname(uri), params=params, headers=__default_headers(auth_token), data=req_body)
    return __handle_response(replace_hostname(uri), response, "PUT")


def do_patch(uri: str, auth_token: str, req_body: str, params=None) -> dict:
    """Perform a PATCH on an endpoint of the YOUnite API"""

    response = requests.patch(replace_hostname(uri), params=params, headers=__default_headers(auth_token), data=req_body)
    return __handle_response(replace_hostname(uri), response, "PATCH")


def do_get(uri: str, auth_token: str, params=None) -> dict:
    """Perform a GET on an endpoint of the YOUnite API"""

    response = requests.get(replace_hostname(uri), params=params, headers=__default_headers(auth_token))
    return __handle_response(replace_hostname(uri), response, "GET")


def do_delete(uri: str, auth_token: str, params=None) -> dict:
    """Perform a DELETE on an endpoint of the YOUnite API"""

    response = requests.delete(replace_hostname(uri), params=params, headers=__default_headers(auth_token))
    return __handle_response(replace_hostname(uri), response, "DELETE")


def get_zone(href: str, auth_token: str, name: str) -> str:
    """Get the UUID of a zone by name. If the zone is not found, a NotFoundError is thrown."""

    data = do_get(href + '/zones', auth_token, {'filters': 'name:' + name})
    zones = data['zones']
    if type(zones) is list and len(zones) > 0:
        uuid = zones[0].get('uuid')
        if type(uuid) is str and len(uuid) > 0:
            return uuid

    raise NotFoundError("ZONE", name)


def get_zone_adaptor(href: str, auth_token: str, zone_uuid: str, adaptor_name: str) -> dict:
    """Get an adaptor by name. If the adaptor is not found, an NotFoundError is thrown."""

    data = do_get(href + '/zones/' + zone_uuid + '/adaptors', auth_token, {'size': '100'})
    adaptors = data.get('adaptors')
    if type(adaptors) is list:
        for adaptor in adaptors:
            if adaptor.get('name') == adaptor_name:
                return adaptor

    raise NotFoundError("ADAPTOR", adaptor_name)


def get_zone_adaptor_registration(href: str, auth_token: str, zone_uuid: str, adaptor_uuid: str) -> dict:
    """
    Get adaptor registration information (client id and secret).
    If the adaptor is not found, an NotFoundError is thrown.
    """

    try:
        return do_get(href + '/zones/' + zone_uuid + '/adaptors/' + adaptor_uuid + '/registration', auth_token)
    except RequestError as e:
        if e.status_code == 404:
            raise NotFoundError("ADAPTOR", adaptor_uuid)
        raise e


def get_domain(href: str, auth_token: str, name: str) -> str:
    """Get the UUID of a domain by name. If not found, NotFoundError is thrown."""

    data = do_get(href + '/domains', auth_token, {'filters': 'name:' + name})
    domains = data['domains']
    if type(domains) is list and len(domains) > 0:
        uuid = domains[0].get("uuid")
        if type(uuid) is str and len(uuid) > 0:
            return uuid

    raise NotFoundError("DOMAIN", name)


def get_domain_version(href: str, auth_token: str, domain_uuid: str, version: int) -> str:
    """Get the UUID of a domain version by domain_uuid and version number. If not found, NotFoundError is thrown."""

    try:
        data = do_get(href + '/domains/' + domain_uuid + '/versions', auth_token, {'filters': 'version:' + str(version)})

        domains = data['domains']
        if type(domains) is list and len(domains) > 0:
            uuid = domains[0].get("uuid")
            if type(uuid) is str and len(uuid) > 0:
                return uuid
    except RequestError as e:
        if e.status_code != 404:
            raise e

    raise NotFoundError("DOMAIN_VERSION", str(version))


def post_zone(href: str, auth_token: str, zone: dict) -> str:
    """
    Create a new zone and return its uuid. If the zone already exists, its uuid will be returned. 
    If the parent zone cannot be found NotFoundError is thrown.
    """
    
    parent_zone_uuid = get_zone(href, auth_token, zone.get('parentZone'))

    uuid = None
    try:
        uuid = get_zone(href, auth_token, zone.get('name'))
    except NotFoundError:
        pass
    
    if uuid is None:
        req_set = {
            "name": zone.get('name'),
            "description": zone.get('description'),
            "parentZoneUuid": parent_zone_uuid,
            "longitude": zone.get('longitude'),
            "latitude": zone.get('latitude'),
            "zoneAdminSsoId": zone.get('zoneAdminSsoId'),
            "zoneDataStewardSsoId": zone.get('zoneDataStewardSsoId')
        }
        
        req_body = json.dumps(req_set)
        response = do_post(href + '/zones', auth_token, req_body)
        uuid = response.get('uuid')
        print("Created zone: " + zone.get('name') + " (" + uuid + ")\n")
    else:
        print("Using existing zone: " + zone.get('name') + " (" + uuid + ")\n")

    return uuid


def post_domain(href: str, auth_token: str, domain: dict) -> str:
    """
    Create a new domain and version and return the uuid of the domain that was created.
    If the domain already exists, its uuid will be returned. 
    If the parent zone cannot be found NotFoundError is thrown.
    """

    uuid = None
    try:
        uuid = get_domain(href, auth_token, domain["name"])
    except NotFoundError:
        pass
    
    if uuid is None:
        # Create the domain
        req_set = {
            "name": domain["name"],
            "zoneUuid": domain["zoneUuid"],
            "domainType": domain["domainType"]
        }
        
        req_body = json.dumps(req_set)
        response = do_post(href + '/domains', auth_token, req_body)
        uuid = response.get('uuid')
        print("Created domain: " + domain.get('name') + " (" + uuid + ")")
    else:
        print("Using existing domain: " + domain.get('name') + " (" + uuid + ")")

    version = "1"
    version_uuid = None
    try:
        version_uuid = get_domain_version(href, auth_token, uuid, version)
    except NotFoundError:
        pass
    
    if version_uuid is None:
        # Create the domain version
        req_set = {
            "modelSchema": domain["modelSchema"],
            "drKeyProperties": domain["drKeyProperties"]
        }

        if "matchingAlgorithm" in domain:
            req_set["matchingAlgorithm"] = domain["matchingAlgorithm"]

        req_body = json.dumps(req_set)
        response = do_post(href + '/domains/' + uuid + '/versions', auth_token, req_body)
        version_uuid = response.get('uuid')
        print("Created domain version: " + domain.get('name') + ": " + version + " (" + version_uuid + ")\n")
    else:
        print("Using existing domain version: " + version + " (" + version_uuid + ")\n")

    return uuid


def patch_domain(href: str, auth_token: str, name: str, version: int, patch: dict) -> str:
    """
    PATCH a domain version
    If the domain version cannot be found NotFoundError is thrown.
    """

    domain_uuid = get_domain(href, auth_token, name)
    version_uuid = get_domain_version(href, auth_token, domain_uuid, version)
    url = href + '/domains/' + domain_uuid + '/versions/' + version_uuid

    print("Patching domain version: " + name + ": " + str(version) + " (" + version_uuid + ")\n")

    data = do_get(url, auth_token)
    patch["changeVersion"] = data["changeVersion"]

    req_body = json.dumps(patch)
    do_patch(url, auth_token, req_body)

    return version_uuid


def post_adaptor(href: str, auth_token: str, adaptor: dict, capabilities: dict, metadata: dict,
                 adaptor_type_uuid: str) -> str:
    """
    Create or update an adaptor by name and return its uuid.
    If the zone does not exist, NotFoundError will be thrown.
    """

    zone_name = adaptor.get('adaptor_zone')
    zone_uuid = get_zone(href, auth_token, zone_name)
    existing = None
    try:
        existing = get_zone_adaptor(href, auth_token, zone_uuid, adaptor.get('adaptor_name'))
    except NotFoundError:
        pass

    req_set = {
        "name": adaptor["adaptor_name"],
        "description": adaptor["adaptor_name"] + " description.",
        "zoneUuid": zone_uuid,
        "capabilities": capabilities,
        "metadata": metadata,
        "adaptorTypeUuid": adaptor_type_uuid
    }

    if existing is None:
        existing = do_post(href + '/zones/' + zone_uuid + '/adaptors', auth_token, json.dumps(req_set))
        adaptor_uuid = existing["uuid"]
        print("Created adaptor: " + adaptor.get('adaptor_name') + " (" + adaptor_uuid + ")\n")
    else:
        adaptor_uuid = existing["uuid"]

        # if the adaptor is running it must be paused first
        started = existing["state"] == "STARTED"
        if started:
            existing = do_patch(href + '/zones/' + zone_uuid + '/adaptors/' + adaptor_uuid, auth_token, json.dumps({
                "changeVersion": existing["changeVersion"],
                "state": "PAUSE"
            }))

        req_set["changeVersion"] = existing["changeVersion"]
        existing = do_put(href + '/zones/' + zone_uuid + '/adaptors/' + adaptor_uuid, auth_token, json.dumps(req_set))

        # if the adaptor was running before, start it back up
        if started:
            do_patch(href + '/zones/' + zone_uuid + '/adaptors/' + adaptor_uuid, auth_token, json.dumps({
                "changeVersion": existing["changeVersion"],
                "state": "PLAY"
            }))

        print("Updated adaptor: " + adaptor.get('adaptor_name') + " (" + adaptor_uuid + ")\n")

    return adaptor_uuid


def post_data_records(href: str, auth_token: str, domain_name: str, domain_version: int, drs: list):
    """Post one or more data records"""

    for dr in drs:
        req_set = {
            "name": domain_name,
            "version": domain_version,
            "json": dr
        }

        req_body = json.dumps(req_set)

        do_post(href + '/drs', auth_token, req_body)


def get_outbound_acls(href: str, auth_token: str, zone_uuid: str) -> dict:
    """Get the first 100 outbound ACLs in a zone. NotFoundError is thrown if the zone is not found."""

    try:
        return do_get(href + '/zones/' + zone_uuid + '/outbound-acls', auth_token, {'size': '100'})
    except RequestError as e:
        if e.status_code == 404:
            raise NotFoundError("ZONE", zone_uuid)
        raise e


def delete_outbound_acl(href: str, auth_token: str, zone_uuid: str, acl_uuid: str):
    """Delete an ACL. NotFoundError is thrown if the zone or ACL is not found."""

    try:
        do_delete(href + '/zones/' + zone_uuid + '/outbound-acls/' + acl_uuid, auth_token)
    except RequestError as e:
        if e.status_code == 404:
            raise NotFoundError("OUTBOUND_ACL", acl_uuid)
        raise e


def post_outbound_acl(href, auth_token, outbound_zone_uuid, outbound_adaptor_uuid, domain_version_uuid,
                      destination_zone_uuid, destination_adaptor_uuid, restrict_get, restrict_put, restrict_post,
                      restrict_delete) -> dict:
    """Create an ACL. NotFoundError is thrown if the zone is not found. The created ACL is returned."""

    # Create the request body
    req_set = {
        "outboundAclCreate": [{
            "sourceAdaptorUuid": outbound_adaptor_uuid,
            "domainVersionUuid": domain_version_uuid,
            "destinationZoneUuid": destination_zone_uuid,
            "destinationAdaptorUuid": destination_adaptor_uuid,
            "restrictGet": restrict_get,
            "restrictPut": restrict_put,
            "restrictPost": restrict_post,
            "restrictDelete": restrict_delete
        }]
    }

    req_body = json.dumps(req_set)
    response = do_post(href + '/zones/' + outbound_zone_uuid + '/outbound-acls', auth_token, req_body)
    print("POSTed Outbound ACL ")
    return response


def get_employee_drs(href: str, auth_token: str, adaptor: str) -> dict:
    """Get employee DRs at an adaptor. """

    data = do_get(href + '/drs', auth_token, {'filters': 'name:' + 'employee,version:1,adaptor:' + adaptor,
                                       'page': 0, 'size': 100})
    employees = data.get('drs')
    employee_map = {}
    if type(employees) is list:
        for employee in employees:
            props = employee.get('drKey')
            employee_id = props.get('employeeId')
            if type(employee_id) is str and len(employee_id) > 0:
                employee_map[employee_id] = employee.get('uuid')
    return employee_map


def get_employee_classification_drs(href: str, auth_token: str) -> dict:
    """Get employee classification DRs."""

    data = do_get(href + '/drs', auth_token, {'filters': 'name:' + 'employeeClassification,version:1'})
    employee_classifications = data.get('drs')
    employee_classification_map = {}
    if type(employee_classifications) is list:
        for employee_classification in employee_classifications:
            props = employee_classification.get('json')
            code = props.get('code')
            if type(code) is str and len(code) > 0:
                employee_classification_map[code] = employee_classification.get('uuid')
    return employee_classification_map


def get_department_drs(href: str, auth_token: str, adaptor: str) -> dict:
    """Get department DRs at an adaptor."""

    data = do_get(href + '/drs', auth_token, {'filters': 'name:' + 'department,version:1,adaptor:' + adaptor,
                                              'page': 0, 'size': 100})
    departments = data.get('drs')
    department_map = {}
    if type(departments) is list:
        for department in departments:
            props = department.get('drKey')
            department_id = props.get('departmentId')
            if type(department_id) is str and len(department_id) > 0:
                department_map[department_id] = department.get('uuid')
    return department_map


def link_employee_drs(href: str, auth_token: str, zone_uuid: str, adaptor_uuid: str, employee_map: dict,
                      employee_classification_map: dict, department_map: dict):

    # For each employee...
    for employee in employee_map:
        dr_uuid = employee_map[employee]
        dr_payload = get_dr_assembler(href, auth_token, dr_uuid, zone_uuid, adaptor_uuid)
        dr_json = json.loads(dr_payload)
        classification_code = dr_json.get('classificationCode')
        manager_employee_id = dr_json.get('managerEmployeeId')
        primary_department_id = dr_json.get('primaryDepartmentId')
        # print(classification_code)
        # print(manager_employee_id)
        # print(primary_department_id)
        classification_code_uuid = employee_classification_map.get(classification_code)
        manager_employee_uuid = employee_map.get(manager_employee_id)
        primary_department_uuid = department_map.get(primary_department_id)
        # print(classification_code_uuid)
        # print(manager_employee_uuid)
        # print(primary_department_uuid)

        uri = href + '/drs/' + dr_uuid + "/link"

        # Make the primaryDepartment $ref

        req_set = {
            "property": "primaryDepartment",
            "uuid": primary_department_uuid
        }
        req_body = json.dumps(req_set)
        do_put(uri, auth_token, req_body)

        # Make the classification $ref
        req_set = {
            "property": "classification",
            "uuid": classification_code_uuid
        }
        req_body = json.dumps(req_set)
        requests.put(replace_hostname(uri), auth_token, req_body)

        if manager_employee_uuid:
            # Make the manager $ref
            req_set = {
                "property": "manager",
                "uuid": manager_employee_uuid
            }
            req_body = json.dumps(req_set)
            requests.put(replace_hostname(uri), auth_token, req_body)

        print("∞ ", end='', flush=True)
    print()


def get_dr_assembler(href: str, auth_token: str, dr_uuid: str, zone_uuid: str, adaptor_uuid: str) -> str:
    """Assemble a DR, returning its json as a string. If the DR is not found, NotFoundError is thrown."""
    
    try:
        params = {'options': 'adaptors:' + adaptor_uuid + ',zone-uuid:' + zone_uuid}
        response_body = do_get(href + '/drs/' + dr_uuid + "/assembler", auth_token, params)
        return response_body['json']
    except RequestError as e:
        if e.status_code == 404:
            raise NotFoundError("DR", dr_uuid)
        raise e
