#!/usr/bin/python3

import getopt
import os
import sys

import adaptors
import databases
import domains
import drs
import zones
from functions import exec_command

#
# Start up the biz demo either locally or in AWS.
# - Creates (if not present) zones and domains first, then adaptors.
#
# Before running this make sure the YOUnite stack is running (locally or in AWS) and that the younite-api server is ready.
#
# To verify that the younite-api server is running, make a request to the /health endpoint from a browser.
#
# For example, if running locally:
#
# http://local.younite.org:8080/health
#
#
# To stop the demo run stop.py
#

assert sys.version_info >= (3, 7)


def usage():
    print("Usage: start.py [-d | -l] [-z] [-o] [-i] [-a adaptor #] ")
    print("-d = Run in AWS Docker/Kubernetes (cannot use with '-i' option)")
    print("-l = Run in Local Docker")
    print("-z = Create zones only (don't create adaptors)")
    print("-o = Create domains only (don't create adaptors)")
    print("-i = When running locally (-l), use the local younite-db-adaptor Docker image (default is to pull the latest from the repo)")


try:
    opts, _ = getopt.getopt(sys.argv[1:], "dlbhzoi:")
except getopt.GetoptError as err:
    print("ERROR: " + err.msg + "\n")
    usage()
    sys.exit(1)

script_path = os.path.dirname(os.path.realpath(__file__))

global_file = script_path + "/../local_global.json"
namespace = "default"
message_bus_url = "tcp://younite-mb:61616"
demo = False
local = False
zones_only = False
domains_only = False
repo_db_adaptor_image = "160909222919.dkr.ecr.us-west-2.amazonaws.com/younite/younite-db-adaptor:latest"
local_db_adaptor_image = "younite/younite-db-adaptor"
db_adaptor_image = repo_db_adaptor_image  # Default is repo db adaptor

for o, v in opts:
    if o == "-d":
        demo = True
        global_file = script_path + "/../aws_global.json"
        message_bus_url = "tcp://younite-mb:61616"
    elif o == "-l":
        local = True
    elif o == "-z":
        zones_only = True
    elif o == "-o":
        domains_only = True
    elif o == "-i":
        db_adaptor_image = local_db_adaptor_image
    elif o == "-h":
        usage()
        sys.exit(0)

if not (demo ^ local):
    usage()
    sys.exit(2)

if demo and db_adaptor_image == local_db_adaptor_image:
    usage()
    sys.exit(3)

if zones_only:
    zones.create_zones(global_file)

if domains_only:
    domains.create_domains(global_file)
    drs.create_drs(global_file)

if zones_only or domains_only:
    sys.exit(0)

try:
    # stop the adaptors and databases (if running)
    if demo:
        adaptors.stop_adaptors_aws_demo(namespace)
        databases.stop_databases_aws_demo(namespace)
    else:
        adaptors.stop_adaptors_local(local_db_adaptor_image)
        adaptors.stop_adaptors_local(repo_db_adaptor_image)
        databases.stop_databases_local()

    # start the databases
    if demo:
        databases.start_databases_aws_demo(namespace)
    else:
        databases.start_databases_local()

    # while databases are starting, create the zones, domains and adaptors
    zones.create_zones(global_file)
    domains.create_domains(global_file)
    drs.create_drs(global_file)
    adaptor_creds = adaptors.create_adaptors(global_file, None)

    # initialize the databases
    if demo:
        databases.init_aws_demo(namespace)
    else:
        databases.init_local(None)

    # start the adaptors
    if demo:
        adaptors.start_adaptors_aws_demo(namespace, adaptor_creds)
    else:
        if db_adaptor_image == repo_db_adaptor_image:
            exec_command(["docker", "pull", repo_db_adaptor_image])

        adaptors.start_adaptors_local(db_adaptor_image, message_bus_url, adaptor_creds)

finally:
    # remove temp files
    for f in os.listdir("."):
        if f.startswith("credentials-") and f.endswith(".json"):
            os.remove(f)

print("Done")
sys.exit(0)
